import * as PIXI from "../pixi/pixi.mjs";
import { loadGameAssets } from "./AssetsLoader/AssetsLoader";
import { UICameraInfo, UIGameProfilerPipe, UIMainPipe, UIObtainedResourcesPipe, UISelectionInfo } from "./UIPipes/UIPipes";
import { getSpriteFromAtlas } from "./Utils/Sprites.utils";
import { profileFPS } from "./Profiler/Profiler";
import { createKeyboardBinding, inputControllerTick } from "./InputController/InputController";
import { BC_APP, BC_BUILDING_PLACEHOLDERS, BC_CAMERA, BC_CURRENT_SCENE, BC_SPRITES_SETTINGS, BC_VIEWPORT, BC_WORLD, PRNG, setBC_APP, setBC_CURRENT_SCENE, setBC_NPC_LAYER, setBC_SELECTION, setBC_VIEWPORT, setBC_WORLD } from "./GlobalVariables/GlobalVariables";
import { clampNumber, interpolate,  } from "./Utils/Math.utils";
import { calculateViewportFromCamera, moveHorizontally, moveVertically, screenToWorldCoordinates } from "./Camera/Camera";

import { handleBuildingsIncome, incBuildingCount } from "./Buildings/Buildings";
import { handleDayNightCycle } from "./World/DayNightCycle";
import { ambientDay, ambientMusic, ambientNight, handleSounds } from "./Sound/Sound";
import { gameStateObjectsCleaner } from "./GameState/GameState";
import { tickHandler } from "./TickHandler/TickHandler";
import { GameScene } from "./GameScene/GameScene";
import { createFirstWorldChunks, getWorldChunkAt } from "./WorldGeneration/WorldGen";
import { ChunkStorageTypes } from "./WorldGeneration/WorldChunk/WorldGenChunk";
import { WorldChunksVisibilityUpdater } from "./WorldGeneration/WorldChunksVisibilityUpdater/WorldChunksVisibilityUpdater";

export function generateWorld() {

    // createFirstWorldChunks();
    startGame();
}

function setupGlobalInput() {
    window.addEventListener("wheel", (event) => {
        const delta = Math.sign(event.deltaY);
        if (delta > 0) {
            BC_CAMERA.targetZoom = Number(clampNumber(BC_CAMERA.targetZoom - BC_CAMERA.zoomStep, BC_CAMERA.minZoom, BC_CAMERA.maxZoom).toFixed(2));
        } else {
            BC_CAMERA.targetZoom = Number(clampNumber(BC_CAMERA.targetZoom + BC_CAMERA.zoomStep, BC_CAMERA.minZoom, BC_CAMERA.maxZoom).toFixed(2));
        }
        UICameraInfo.update((s) => {
            s.zoom = BC_CAMERA.targetZoom;
        });
    });
}

function setupInGameSelector() {

    let sprite0 = PIXI.Sprite.from("assets/images/selection.png");
    sprite0.scale.set(BC_SPRITES_SETTINGS.scale, BC_SPRITES_SETTINGS.scale);
    BC_VIEWPORT.addChild(sprite0);
    sprite0.zIndex = 100;
    setBC_SELECTION(sprite0);

    BC_VIEWPORT.eventMode = "static";
    BC_VIEWPORT.onmousemove = (e) => {
        let t = screenToWorldCoordinates(e.data.global.x, e.data.global.y);
        // let tChunk = getWorldChunkAt(t.x, t.y);
        // let tTerrainTile = tChunk.getFromChunk(t.x, t.y, ChunkStorageTypes.TYPE_TERRAIN);

        t.x = BC_SPRITES_SETTINGS.defaultSize * BC_SPRITES_SETTINGS.scale * Math.floor(t.x / (BC_SPRITES_SETTINGS.defaultSize * BC_SPRITES_SETTINGS.scale));
        t.y = BC_SPRITES_SETTINGS.defaultSize * BC_SPRITES_SETTINGS.scale * Math.floor(t.y / (BC_SPRITES_SETTINGS.defaultSize * BC_SPRITES_SETTINGS.scale));
        sprite0.position.set(t.x, t.y);
    };
}

export async function initGame() {
    const app = new PIXI.Application();
    await app.init({ resizeTo: window, antialias: false });
    document.body.appendChild(app.canvas);
    setBC_APP(app);
    app.ticker.targetFPMS = 0.14;
    app.ticker.maxFPS = 140;

    await loadGameAssets();
    UIMainPipe.update((s) => {
        s.loadingAssets = false;
    });

    // let world = new PIXI.Container();
    let viewport = new PIXI.Container();
    // let NPCLayer = new PIXI.Container();
    // world.addChild(viewport);
    // viewport.addChild(NPCLayer);
    setBC_VIEWPORT(viewport);
    // setBC_WORLD(world);
    // setBC_NPC_LAYER(NPCLayer);
    // NPCLayer.zIndex = 100;
    viewport.isRenderGroup = true;
    // NPCLayer.isRenderGroup = true;
    app.stage.addChild(viewport);
    // world.tint = 0x00ffff;

    setBC_CURRENT_SCENE(new GameScene());

    BC_CAMERA.position.x = Math.floor(PRNG() * 3242 - 372);
    BC_CAMERA.position.y = Math.floor(PRNG() * 1285 - 255);

    UICameraInfo.update((s)=>{
        s.position.x = BC_CAMERA.position.x;
        s.position.y = BC_CAMERA.position.y;
    })

    app.ticker.add(inputControllerTick);
    app.ticker.add(calculateViewportFromCamera);
    // console.log(WorldChunksVisibilityUpdater.handleWorldChunksVisibility);
    app.ticker.add(()=>{WorldChunksVisibilityUpdater.handleWorldChunksVisibility();});
    app.ticker.add(()=>{WorldChunksVisibilityUpdater.handleWorldChunkFilling();});
    app.ticker.add(profileFPS);
    app.ticker.add(handleBuildingsIncome);

    createKeyboardBinding("KeyW", "Hold", [moveVertically]);
    createKeyboardBinding("KeyS", "Hold", [moveVertically]);
    createKeyboardBinding("KeyA", "Hold", [moveHorizontally]);
    createKeyboardBinding("KeyD", "Hold", [moveHorizontally]);
}

function startGame() {
    setupGlobalInput();
    setupInGameSelector();
    BC_APP.ticker.add(handleDayNightCycle);
    BC_APP.ticker.add(handleSounds);
    ambientDay.play();
    ambientNight.play();
    ambientMusic.play();
    // addNPCToWorld(BC_CAMERA.position.x, BC_CAMERA.position.y, {type: "slave"});
    BC_APP.ticker.add(gameStateObjectsCleaner);
    BC_APP.ticker.add(tickHandler);

    createFirstWorldChunks();
}
